from unittest.mock import Mock

import pytest
from celery import Celery
from celery.contrib.testing.worker import TestWorkController

from acoupi import data
from acoupi.components import SqliteStore
from acoupi.components.audio_recorder import MicrophoneConfig
from acoupi.components.types import Model
from acoupi.programs.templates import (
    AudioConfiguration,
    DetectionProgram,
    DetectionProgramConfiguration,
    DetectionsConfiguration,
    MessagingConfig,
    PathsConfiguration,
)


class Config(DetectionProgramConfiguration):
    name: str = "Test Program"


@pytest.fixture
def detection_config() -> DetectionsConfiguration:
    return DetectionsConfiguration(threshold=0.4)


@pytest.fixture
def config(
    audio_config: AudioConfiguration,
    paths_config: PathsConfiguration,
    microphone_config: MicrophoneConfig,
    messaging_config: MessagingConfig,
    detection_config: DetectionsConfiguration,
) -> Config:
    return Config(
        name="test",
        paths=paths_config,
        microphone=microphone_config,
        messaging=messaging_config,
        recording=audio_config,
        detections=detection_config,
    )


@pytest.mark.usefixtures("celery_app")
def test_can_initialise_program(
    celery_app: Celery,
    config: Config,
):
    model = Mock()

    class Program(DetectionProgram):
        config_schema = Config

        def configure_model(self, config):
            return model

    Program(config, celery_app)


@pytest.mark.usefixtures("celery_app")
def test_program_check_runs_all_relevant_checks(
    celery_app: Celery,
    config: Config,
):
    model = Mock()
    messenger = Mock()
    recorder = Mock()

    class Program(DetectionProgram):
        config_schema = Config

        def configure_model(self, config):
            return model

        def configure_messenger(self, config):
            return messenger

        def configure_recorder(self, config):
            return recorder

    program = Program(config, celery_app)

    program.check(config)
    model.check.assert_called_once()
    recorder.check.assert_called_once()
    messenger.check.assert_called_once()


@pytest.mark.usefixtures("celery_app")
def test_program_has_all_required_tasks(
    celery_app: Celery,
    config: Config,
):
    model = Mock()

    class Program(DetectionProgram):
        config_schema = Config

        def configure_model(self, config):
            return model

    program = Program(config, celery_app)

    assert "recording_task" in program.tasks
    assert "file_management_task" in program.tasks
    assert "detection_task" in program.tasks
    assert "heartbeat_task" in program.tasks
    assert "send_messages_task" in program.tasks


class MockModel(Model):
    name: str = "test_model"

    def run(self, recording: data.Recording) -> data.ModelOutput:
        return data.ModelOutput(
            name_model="test",
            recording=recording,
            detections=[
                data.PresenceDetection(
                    detection_score=score / 10,
                    tags=[
                        data.PredictedTag(
                            tag=data.Tag(
                                key="species", value="Myotis daubentonii"
                            ),
                            confidence_score=score2 / 10,
                        )
                        for score2 in range(11)
                    ],
                )
                for score in range(11)
            ],
        )


def test_worker_config_has_dedicated_detection_worker():
    worker_config = DetectionProgram.worker_config
    assert worker_config is not None
    workers = {w.name: w for w in worker_config.workers}
    assert "detection" in workers, (
        "DetectionProgram must declare a 'detection' worker"
    )
    assert workers["detection"].concurrency == 1, (
        "detection worker must have concurrency=1 to prevent OOM from parallel model loads"
    )
    assert workers["recording"].concurrency == 1, (
        "recording worker must have concurrency=1 (audio device is not thread-safe)"
    )


@pytest.mark.usefixtures("celery_app")
def test_detection_task_is_routed_to_detection_queue(
    celery_app: Celery,
    config: Config,
):
    class Program(DetectionProgram):
        config_schema = Config

        def configure_model(self, config):
            return Mock()

    Program(config, celery_app)

    routes = celery_app.conf.task_routes or {}
    assert routes.get("detection_task") == {"queue": "detection"}, (
        "detection_task must be routed to the 'detection' queue so that "
        "apply_async() dispatches it to the concurrency=1 worker"
    )


@pytest.mark.usefixtures("celery_app")
def test_unknown_callback_queue_raises_at_setup(
    celery_app: Celery,
    config: Config,
):
    """Passing a callback_queue not in worker_config must fail early."""

    class Program(DetectionProgram):
        config_schema = Config

        def configure_model(self, config):
            return Mock()

        def register_recording_task(self, config):
            recording_task = self.create_recording_task(config)
            self.add_task(
                function=recording_task,
                callbacks=self.get_recording_callbacks(config),
                queue="recording",
                callback_queue="nonexistent_queue",
            )

    with pytest.raises(ValueError, match="nonexistent_queue"):
        Program(config, celery_app)


def test_program_only_stores_detections_with_high_threshold(
    celery_app: Celery,
    celery_worker: TestWorkController,
    config: Config,
    recording: data.Recording,
):
    threshold = 0.6

    class Program(DetectionProgram):
        config_schema = Config

        def configure_model(self, config):
            return MockModel()

    config = config.model_copy(
        update=dict(detections=DetectionsConfiguration(threshold=threshold))
    )

    program = Program(config, celery_app)
    celery_worker.reload()

    program.store.store_recording(recording)

    assert "detection_task" in program.tasks
    program.tasks["detection_task"].delay(recording).get()

    assert isinstance(program.store, SqliteStore)

    outputs = program.store.get_model_outputs(recording_ids=[recording.id])
    assert len(outputs) == 1
    output = outputs[0]

    # Check all scores are above threshold
    assert all(
        detection.detection_score >= threshold
        for detection in output.detections
    )
    assert all(
        tag.confidence_score >= threshold
        for detection in output.detections
        for tag in detection.tags
    )

    # Check no detection/tag is missed
    assert (
        min(detection.detection_score for detection in output.detections)
        == threshold
    )
    assert (
        min(
            tag.confidence_score
            for detection in output.detections
            for tag in detection.tags
        )
        == threshold
    )



"""Test suite for acoupi messengers."""

from pathlib import Path
from unittest import mock

import pytest
import pytest_httpserver
import requests
from paho.mqtt.enums import MQTTErrorCode

from acoupi import data
from acoupi.components import messengers
from acoupi.system.exceptions import HealthCheckError, MessageSendError


def test_http_messenger():
    """Test the HTTPMessenger."""
    with mock.patch("requests.post") as mock_post:
        mock_post.return_value.status_code = 200
        mock_post.return_value.text = "OK"

        messenger = messengers.HTTPMessenger(
            base_url="http://localhost:5000",
            timeout=5,
        )

        message = data.Message(content='"Hello, world!"')

        response = messenger.send_message(message)

        assert response.status == data.ResponseStatus.SUCCESS
        assert response.content == "OK"

        mock_post.assert_called_once_with(
            "http://localhost:5000",
            json="Hello, world!",
            params={},
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
            timeout=5,
        )


def test_http_messenger_with_params():
    """Test the HTTPMessenger with parameters."""
    with mock.patch("requests.post") as mock_post:
        mock_post.return_value.status_code = 200
        mock_post.return_value.text = "OK"

        messenger = messengers.HTTPMessenger(
            base_url="http://localhost:5000",
            base_params={"key": "value"},
            timeout=5,
        )

        message = data.Message(content='"Hello, world!"')

        response = messenger.send_message(message)

        assert response.status == data.ResponseStatus.SUCCESS
        assert response.content == "OK"

        mock_post.assert_called_once_with(
            "http://localhost:5000",
            json="Hello, world!",
            params={"key": "value"},
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
            timeout=5,
        )


def test_http_messenger_accepts_utf8_json_bytes():
    """Test the HTTPMessenger accepts JSON payloads encoded as bytes."""
    with mock.patch("requests.post") as mock_post:
        mock_post.return_value.status_code = 200
        mock_post.return_value.text = "OK"

        messenger = messengers.HTTPMessenger(
            base_url="http://localhost:5000",
            timeout=5,
        )

        message = data.Message(content=b'"Hello, world!"')

        response = messenger.send_message(message)

        assert response.status == data.ResponseStatus.SUCCESS
        assert response.content == "OK"

        mock_post.assert_called_once_with(
            "http://localhost:5000",
            json="Hello, world!",
            params={},
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
            timeout=5,
        )


def test_http_messenger_rejects_non_utf8_bytes():
    """Test the HTTPMessenger rejects bytes that are not UTF-8."""
    with mock.patch("requests.post") as mock_post:
        messenger = messengers.HTTPMessenger(
            base_url="http://localhost:5000",
            timeout=5,
        )

        message = data.Message(content=b"\x80\x81\x82")

        with pytest.raises(
            MessageSendError, match="Invalid HTTP message payload"
        ):
            messenger.send_message(message)

        mock_post.assert_not_called()


def test_http_messenger_rejects_invalid_json_bytes():
    """Test the HTTPMessenger rejects UTF-8 bytes that are not JSON."""
    with mock.patch("requests.post") as mock_post:
        messenger = messengers.HTTPMessenger(
            base_url="http://localhost:5000",
            timeout=5,
        )

        message = data.Message(content=b"not json")

        with pytest.raises(
            MessageSendError, match="Invalid HTTP message payload"
        ):
            messenger.send_message(message)

        mock_post.assert_not_called()


def test_http_messenger_raises_on_request_exception():
    """Test the HTTPMessenger raises on local request failures."""
    with mock.patch("requests.post") as mock_post:
        mock_post.side_effect = requests.exceptions.ConnectionError("boom")
        messenger = messengers.HTTPMessenger(
            base_url="http://localhost:5000",
            timeout=5,
        )

        message = data.Message(content='"Hello, world!"')

        with pytest.raises(MessageSendError, match="HTTP request failed"):
            messenger.send_message(message)


def test_mqtt_messenger_raises_when_connection_fails():
    """Test the MQTTMessenger raises on local connection failures."""
    with mock.patch("paho.mqtt.client.Client") as mock_client:
        mock_client.return_value.is_connected.return_value = False
        mock_client.return_value.connect.return_value = (
            MQTTErrorCode.MQTT_ERR_NO_CONN
        )
        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )

        message = data.Message(content='"Hello, world!"')

        with pytest.raises(MessageSendError, match="MQTT connection error"):
            messenger.send_message(message)


def test_http_messeger_with_custom_headers():
    """Test the HTTPMessenger with custom headers."""
    with mock.patch("requests.post") as mock_post:
        mock_post.return_value.status_code = 200
        mock_post.return_value.text = "OK"

        messenger = messengers.HTTPMessenger(
            base_url="http://localhost:5000",
            headers={"X-Test": "value"},
            timeout=5,
        )

        message = data.Message(content='"Hello, world!"')

        response = messenger.send_message(message)

        assert response.status == data.ResponseStatus.SUCCESS
        assert response.content == "OK"

        mock_post.assert_called_once_with(
            "http://localhost:5000",
            json="Hello, world!",
            params={},
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
                "X-Test": "value",
            },
            timeout=5,
        )


def test_http_messenger_with_complex_message():
    """Test the HTTPMessenger with a complex message."""
    with mock.patch("requests.post") as mock_post:
        mock_post.return_value.status_code = 200
        mock_post.return_value.text = "OK"

        messenger = messengers.HTTPMessenger(
            base_url="http://localhost:5000",
            timeout=5,
        )

        model_output = data.ModelOutput(
            name_model="test_model",
            recording=data.Recording(
                path=Path("test_recording.wav"),
                duration=1,
                samplerate=16000,
                created_on=data.utc_now(),
                deployment=data.Deployment(
                    name="test_deployment",
                ),
            ),
            detections=[
                data.PresenceDetection(
                    location=data.BoundingBox(
                        coordinates=(0, 0, 1, 1),
                    ),
                    detection_score=0.5,
                    tags=[
                        data.PredictedTag(
                            tag=data.Tag(
                                key="test_tag",
                                value="test_value",
                            ),
                            confidence_score=0.2,
                        ),
                        data.PredictedTag(
                            tag=data.Tag(
                                key="event",
                                value="echolocation",
                            ),
                            confidence_score=0.8,
                        ),
                    ],
                ),
            ],
        )

        message = data.Message(content=model_output.model_dump_json())

        response = messenger.send_message(message)

        assert response.status == data.ResponseStatus.SUCCESS
        assert response.content == "OK"

        mock_post.assert_called_once_with(
            "http://localhost:5000",
            json=model_output.model_dump(mode="json"),
            params={},
            headers={
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
            timeout=5,
        )


def test_http_messenger_is_sending_a_message(
    httpserver: pytest_httpserver.HTTPServer,
):
    """Test the HTTPMessenger is sending a message."""
    # Arrange
    httpserver.expect_request("/endpoint").respond_with_data("OK")
    messenger = messengers.HTTPMessenger(
        base_url=httpserver.url_for("/endpoint"),
        timeout=5,
    )
    message = data.Message(content='"Hello, world!"')

    # Act
    response = messenger.send_message(message)

    # Assert
    assert response.status == data.ResponseStatus.SUCCESS
    assert response.content == "OK"


def test_http_messenger_fails_with_bad_request(
    httpserver: pytest_httpserver.HTTPServer,
):
    """Test the HTTPMessenger fails with a bad request."""
    # Arrange
    httpserver.expect_request("/endpoint").respond_with_data(
        "BAD REQUEST",
        status=400,
    )
    messenger = messengers.HTTPMessenger(
        base_url=httpserver.url_for("/endpoint"),
        timeout=5,
    )
    message = data.Message(content='"Hello, world!"')

    # Act
    response = messenger.send_message(message)

    # Assert
    assert response.status == data.ResponseStatus.ERROR
    assert response.content == "BAD REQUEST"


def test_http_check_is_successful(httpserver: pytest_httpserver.HTTPServer):
    """Test the HTTPMessenger check connection is successful."""
    # Arrange
    httpserver.expect_request("/", method="OPTIONS").respond_with_data("OK")
    messenger = messengers.HTTPMessenger(
        base_url=httpserver.url_for("/"),
        timeout=5,
    )

    # Act
    messenger.check()


def test_http_check_fails_is_cannot_connect():
    """Test the HTTPMessenger check connection fails if cannot connect."""
    # Arrange
    messenger = messengers.HTTPMessenger(
        base_url="http://localhost:5000",
        timeout=5,
    )

    # Act
    with pytest.raises(HealthCheckError):
        messenger.check()


def test_http_check_fails_is_not_allowed(
    httpserver: pytest_httpserver.HTTPServer,
):
    """Test the HTTPMessenger check connection fails if not allowed."""
    # Arrange
    httpserver.expect_request("/", method="OPTIONS").respond_with_data(
        "NOT ALLOWED",
        status=405,
    )
    messenger = messengers.HTTPMessenger(
        base_url=httpserver.url_for("/"),
        timeout=5,
    )

    # Act
    with pytest.raises(HealthCheckError, match="405"):
        messenger.check()


def test_http_check_fails_if_post_not_allowed(
    httpserver: pytest_httpserver.HTTPServer,
):
    """Test the HTTPMessenger check connection fails if POST not allowed."""
    # Arrange
    httpserver.expect_request("/", method="OPTIONS").respond_with_data(
        "OK", headers={"Allow": "GET"}
    )
    messenger = messengers.HTTPMessenger(
        base_url=httpserver.url_for("/"),
        timeout=5,
    )

    # Act
    with pytest.raises(HealthCheckError, match="POST"):
        messenger.check()


def test_mqtt_messenger_can_send_simple_message():
    """Test the MQTTMessenger can send a simple message."""
    # Arrange
    mock_response = mock.MagicMock()
    mock_response.rc = 0

    with mock.patch(
        "paho.mqtt.client.Client",
        spec=True,
    ) as mock_client:
        mock_client.return_value.is_connected.return_value = True
        mock_client.return_value.publish.return_value = mock_response
        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )
        message = data.Message(content='"Hello, world!"')

        # Act
        response = messenger.send_message(message)

        # Assert
        assert response.status == data.ResponseStatus.SUCCESS
        assert response.content == "MQTT_ERR_SUCCESS"

        mock_client.return_value.publish.assert_called_once_with(
            topic="acoupi",
            payload='"Hello, world!"',
        )


def test_mqtt_messenger_can_send_bytes_message():
    """Test the MQTTMessenger can send a bytes payload."""
    mock_response = mock.MagicMock()
    mock_response.rc = 0

    with mock.patch(
        "paho.mqtt.client.Client",
        spec=True,
    ) as mock_client:
        mock_client.return_value.is_connected.return_value = True
        mock_client.return_value.publish.return_value = mock_response
        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )
        message = data.Message(content=b"\x01\x02payload")

        response = messenger.send_message(message)

        assert response.status == data.ResponseStatus.SUCCESS
        assert response.content == "MQTT_ERR_SUCCESS"

        mock_client.return_value.publish.assert_called_once_with(
            topic="acoupi",
            payload=b"\x01\x02payload",
        )


def test_mqtt_messenger_returns_error_response_for_publish_rc_error():
    """Test the MQTTMessenger returns a response for broker publish errors."""
    mock_response = mock.MagicMock()
    mock_response.rc = MQTTErrorCode.MQTT_ERR_QUEUE_SIZE

    with mock.patch(
        "paho.mqtt.client.Client",
        spec=True,
    ) as mock_client:
        mock_client.return_value.is_connected.return_value = True
        mock_client.return_value.publish.return_value = mock_response

        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )
        message = data.Message(content='"Hello, world!"')

        response = messenger.send_message(message)

        assert response.status == data.ResponseStatus.ERROR
        assert response.content == "MQTT_ERR_QUEUE_SIZE"

        mock_client.return_value.publish.assert_called_once_with(
            topic="acoupi",
            payload='"Hello, world!"',
        )


def test_mqtt_check_is_succesful_when_connected():
    """Test the MQTTMessenger check connection is successful when connected."""
    # Arrange
    with mock.patch("paho.mqtt.client.Client") as mock_client:
        mock_client.return_value.is_connected.return_value = True
        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )

        # Act
        messenger.check()


def test_mqtt_check_is_succesful_when_can_reconnect():
    """Test the MQTTMessenger check connection is successful when can reconnect."""
    # Arrange
    called = {"is_conected": False}

    def is_conected():
        if not called["is_conected"]:
            called["is_conected"] = True
            return False
        return True

    with mock.patch(
        "paho.mqtt.client.Client",
        spec=True,
    ) as mock_client:
        mock_client.return_value.is_connected.return_value = True
        mock_client.return_value.reconnect.return_value = 0
        mock_client.return_value.host = "localhost"

        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )

        # Act
        messenger.check()


def test_mqtt_check_is_successful_when_can_connect():
    """Test the MQTTMessenger check connection is successful when can connect."""
    # Arrange
    called = {"is_conected": False}

    def is_conected():
        if not called["is_conected"]:
            called["is_conected"] = True
            return False
        return True

    with mock.patch(
        "paho.mqtt.client.Client",
        spec=True,
    ) as mock_client:
        mock_client.return_value.is_connected.return_value = False
        mock_client.return_value.connect.return_value = 0
        mock_client.return_value.host = None

        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )

        # Act
        messenger.check()


@pytest.mark.parametrize(
    "error_code",
    range(1, 17),
)
def test_mqtt_check_fails_with_mqtt_error_on_connection(error_code: int):
    """Test the MQTTMessenger check connection fails with MQTT error."""
    # Arrange
    with mock.patch(
        "paho.mqtt.client.Client",
        spec=True,
    ) as mock_client:
        mock_client.return_value.is_connected.return_value = False
        mock_client.return_value.connect.return_value = error_code
        mock_client.return_value.host = None

        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )

        # Act
        with pytest.raises(
            HealthCheckError,
            match=MQTTErrorCode(error_code).name,
        ):
            messenger.check()


@pytest.mark.parametrize(
    "error_code",
    range(1, 17),
)
def test_mqtt_check_fails_with_mqtt_error_on_reconnection(error_code: int):
    """Test the MQTTMessenger check connection fails with MQTT error."""
    # Arrange
    with mock.patch(
        "paho.mqtt.client.Client",
        spec=True,
    ) as mock_client:
        mock_client.return_value.is_connected.return_value = False
        mock_client.return_value.reconnect.return_value = error_code
        mock_client.return_value.host = "localhost"

        messenger = messengers.MQTTMessenger(
            host="localhost",
            port=1883,
            username="test",
            topic="acoupi",
        )

        # Act
        with pytest.raises(
            HealthCheckError,
            match=MQTTErrorCode(error_code).name,
        ):
            messenger.check()


def test_mqtt_check_fails_with_bad_host():
    """Test the MQTTMessenger check connection fails if auth failed."""
    # Arrange
    messenger = messengers.MQTTMessenger(
        host="unexistent.mosquitto.org",
        port=1884,
        username="test",
        topic="#",
    )
    # Act
    with pytest.raises(HealthCheckError):
        messenger.check()



"""Processing filters for the Acoupi project.

Processing filters are used to determine if a recording should be processed by
the model. This is useful for example if you want to only process recordings
that satisfy certain criteria, such as a minimum duration or that surpass a
certain amplitude threshold. This can be used to reduce the amount of
computational resources required to process a large number of recordings.

Processing filters are implemented as classes that inherit from
ProcessingFilter. The class should implement the should_process_recording
method, which takes a Recording object and returns a boolean indicating if the
recording should be processed by the model.

Keep in mind that the should_process_recording method is called for every
recording, so it should be as efficient as possible.
"""

from acoupi import data
from acoupi.components.types import ProcessingFilter

__all__ = [
    "TrivialProcessingFilter",
]


class TrivialProcessingFilter(ProcessingFilter):
    """A ProcessingFilter that always returns True."""

    def should_process_recording(self, recording: data.Recording) -> bool:
        """Determine if the recording should be processed by the model."""
        return True



"""CLI commands to manage acoupi configuration."""

import click
import pygments
from pygments.formatters import TerminalFormatter  # type: ignore
from pygments.lexers import JsonLexer  # type: ignore

from acoupi import system
from acoupi.cli.base import acoupi
from acoupi.system import exceptions

__all__ = [
    "config",
]


@acoupi.group()
@click.pass_context
def config(ctx):
    """Manage the acoupi configuration.

    This command group provides subcommands to view and modify the acoupi
    configuration settings.

    Before using any subcommands, ensure that acoupi is properly set up by
    running `acoupi setup`.
    """
    settings = ctx.obj["settings"]

    if not system.is_configured(settings):
        click.echo("Acoupi is not setup. Run `acoupi setup` first.")
        raise click.Abort()

    try:
        schema = system.load_config_schema(settings)
    except exceptions.ProgramNotFoundError as err:
        click.secho(
            f"We couldn't find the program\n\nError: {err.program}",
            fg="red",
        )
        return
    except exceptions.InvalidProgramError as err:
        click.secho(
            f"The program is invalid\n\nError: {err.program}",
            fg="red",
        )
        raise click.Abort() from err

    try:
        config = system.load_config(
            settings.program_config_file,
            schema,
        )
    except exceptions.ConfigurationError as err:
        click.secho(
            f"The configuration file is invalid\n\nError: {err}",
            fg="red",
        )
        raise click.Abort() from err

    ctx.obj["config"] = config


@config.command("get")
@click.pass_context
@click.option(
    "--field",
    type=str,
    help=(
        "Retrieve a specific field or nested field from the configuration "
        "using dot notation (e.g., 'section.subsection.value'). If not "
        "provided, the entire configuration is displayed."
    ),
)
@click.option(
    "--color/--no-color",
    default=True,
    help=(
        "Enable or disable syntax highlighting for improved readability. "
        "Default is enabled."
    ),
)
@click.option(
    "--indent",
    "-i",
    type=int,
    default=2,
    help="Set the indentation level for the output JSON. Default is 2 spaces.",
)
def get_config_field(ctx, field: str, color: bool, indent: int):
    """Display the full (or a specific field of the) acoupi configuration.

    This command allows you to view the current configuration settings for
    acoupi. You can either retrieve the entire configuration or a specific
    field or nested field within the configuration.

    Examples
    --------
    To display the entire configuration:

        acoupi config get

    To display the value of the `username` field:

        acoupi config get --field username

    To display the value of the nested field `server.port`:

        acoupi config get --field server.port

    To display the configuration without color highlighting:

        acoupi config get --no-color
    """
    config = ctx.obj["config"]

    if field:
        config = system.get_config_field(config, field)

    output = system.dump_config(config, indent=indent)

    if color:
        output = pygments.highlight(
            output,
            JsonLexer(),
            TerminalFormatter(),
        )

    click.echo(output)


@config.command("set")
@click.argument("value", required=True, type=str)
@click.option(
    "--field",
    type=str,
    default="",
    help=(
        "Set a specific field of the configuration using dot notation "
        "(e.g., 'section.subsection.value'). "
        "If not provided, the entire configuration is modified."
    ),
)
@click.pass_context
def set_field(ctx, field: str, value: str):
    """Set a specific field or the entire acoupi configuration to a new VALUE.

    This command allows you to update a configuration value by specifying its
    new value. You can:

    * Modify the entire configuration (if `--field` is not provided)

    * Modify a specific field

    * Modify nested fields using dot notation (e.g., 'section.subsection.value')

    Note: Any configuration changes provided are validated against the original
    configuration schema to ensure data integrity.

    Examples
    --------
    To change the `username` to "new_user":

        acoupi config set username new_user

    To change the nested field `server.port` to 8080:

        acoupi config set server.port 8080

    To replace the entire configuration with a new one (assuming
    'new_config.json' contains valid JSON):

        acoupi config set "$(cat new_config.json)"
    """
    config = ctx.obj["config"]
    settings = ctx.obj["settings"]

    try:
        new_config = system.set_config_field(config, field, value)
    except (
        IndexError,
        ValueError,
        AttributeError,
    ) as err:
        raise click.UsageError(
            f"Invalid field or value. \n"
            f"{click.style(err, fg='red', bold=True)}",
            ctx=ctx,
        ) from err

    system.write_config(new_config, settings.program_config_file)



"""PyAudio device discovery helpers for audio input devices."""

from typing import List, Tuple

import pyaudio
from pydantic import BaseModel

from acoupi.system.exceptions import DeviceUnavailableError

__all__ = [
    "get_input_devices",
    "get_input_device_by_name",
    "has_input_audio_device",
    "get_default_microphone",
]


class DeviceInfo(BaseModel):
    """Normalized description of a PyAudio input device."""

    index: int
    """The input index of the audio device."""

    name: str
    """The name of the audio device."""

    max_input_channels: int
    """The maximum number of input channels."""

    default_samplerate: float
    """The default samplerate of the audio device."""


def get_input_devices(p: pyaudio.PyAudio) -> List[DeviceInfo]:
    """Return all PyAudio devices that support audio input."""
    count = p.get_device_count()

    devices = []
    for i in range(count):
        info = p.get_device_info_by_index(i)

        if int(info["maxInputChannels"]) > 0:
            devices.append(
                DeviceInfo.model_validate(
                    {
                        "index": i,
                        "name": parse_device_name(str(info["name"])),
                        "max_input_channels": info["maxInputChannels"],
                        "default_samplerate": info["defaultSampleRate"],
                    }
                )
            )

    return devices


def parse_device_name(full_name: str) -> str:
    """Parse the name of an audio device.

    Notes
    -----
    PyAudio appends additional information to the name of the audio device
    in the form of "<name>: <additional_info>". This function removes the
    additional information and returns the name of the audio device.
    """
    return full_name.split(":")[0].strip()


def get_input_device_by_name(p: pyaudio.PyAudio, name: str) -> DeviceInfo:
    """Get an input device by its name.

    Parameters
    ----------
    p: pyaudio.PyAudio
        An instance of PyAudio.
    name: str
        The name of the audio device.

    Returns
    -------
    DeviceInfo
        The information of the audio device.

    Raises
    ------
    DeviceUnavailableError
        If the named input device cannot be found.

    Notes
    -----
    It is assumed that the name of the audio device has been cleaned
    using the `parse_device_name` function, i.e., the additional information
    added by PyAudio has been removed. This should coincide with the name
    provided by `arecord -l` or `lsusb`.
    """
    available_devices = get_input_devices(p)

    for device in available_devices:
        if name == device.name:
            return device

    raise DeviceUnavailableError(
        message=(
            f"Audio device with name '{name}' not found."
            f" Available devices: {', '.join([device.name for device in available_devices])}"
        )
    )


def has_input_audio_device() -> bool:
    """Return ``True`` when a default PyAudio input device is available."""
    p = pyaudio.PyAudio()

    try:
        p.get_default_input_device_info()
        return True
    except OSError:
        return False


def get_default_microphone() -> Tuple[int, int, str]:
    """Return the default PyAudio input device parameters.

    Returns
    -------
    channels: int
        The number of audio channels of the audio device.
    sample_rate: int
        The samplerate of the audio device.
    device_name: str
        The name of the audio device.

    Raises
    ------
    DeviceUnavailableError
        If no compatible default input device is available.
    """
    # Create an instance of PyAudio
    p = pyaudio.PyAudio()
    try:
        p.get_default_input_device_info()

        # Get default input device info
        default_input_device = p.get_default_input_device_info()

        # Get the number of channels
        channels = int(default_input_device["maxInputChannels"])

        # Get the input device name
        name = str(default_input_device["name"])

        # Get the sample rate
        sample_rate = int(default_input_device["defaultSampleRate"])

        p.terminate()
        return channels, sample_rate, name

    except OSError as error:
        raise DeviceUnavailableError(
            "No compatible audio device found."
        ) from error



"""ModelOutput cleaners for acoupi.

Model Output Cleaners are responsible for cleaning the outputs of a model (i.e.,
detections) that do not meet certain criteria. This can include removing low
confidence tags and detections, or detections and tags that have a specific
labels. For example, the ThresholdDetectionCleaner removes any predictions
(i.e., detections and tags) with a score below a threshold.

The ModelOutputCleaner is implemented as a class that inherits from
ModelOutputCleaner. The class should implement the clean method, which takes a
data.ModelOutput object and returns a cleaned data.ModelOutput object. The
modeloutput that does not meet the criteria are removed.

The ModelOutputCleaner is used in the detection task to clean the outputs of
the model BEFORE storing them in the store. The ModelOutputCleaner is passed to
the detection task as a list of ModelOutputCleaner objects. This allows to use
multiple ModelOutputCleaners to clean the model output.
"""

from typing import List, Sequence

from acoupi import data
from acoupi.components import types


class ThresholdDetectionCleaner(types.ModelOutputCleaner):
    """Keeps predictions with a score higher than a threshold.

    This class implements a model output cleaner that removes any
    predictions with a score below a threshold. This includes
    removing low confidence tags and detections.
    """

    detection_threshold: float
    """The threshold to use to define when a detection is stored."""

    def __init__(self, detection_threshold: float):
        """Initiatlise the filter."""
        self.detection_threshold = detection_threshold

    def clean(self, model_output: data.ModelOutput) -> data.ModelOutput:
        """Clean the model output.

        Parameters
        ----------
        model_output : data.ModelOutput
            The model output to clean.

        Returns
        -------
        data.ModelOutput
            The cleaned model output.

        Examples
        --------
        >>> model_output = data.ModelOutput(
        ...     detections=[
        ...         data.PresenceDetection(
        ...             detection_score=0.8,
        ...             tags=[
        ...                 data.PredictedTag(
        ...                     tag=data.Tag(
        ...                         key="species", value="species_1"
        ...                     ),
        ...                     confidence_score=0.7,
        ...                 )
        ...                 data.PredictedTag(
        ...                     tag=data.Tag(
        ...                         key="species", value="species_2"
        ...                     ),
        ...                     confidence_score=0.4,
        ...                 )
        ...             ],
        ...         )
        ...     ]
        ... )
        >>> cleaner = ThresholdDetectionCleaner(detection_threshold=0.6)
        >>> model_output = cleaner.clean(model_output)
        >>> assert model_output == data.ModelOutput(
        ...     detections=[
        ...         data.PresenceDetection(
        ...             detection_score=0.8,
        ...             tags=[
        ...                 data.PredictedTag(
        ...                     tag=data.Tag(
        ...                         key="species", value="species_1"
        ...                     ),
        ...                     confidence_score=0.7,
        ...                 )
        ...             ],
        ...         )
        ...     ]
        ... )
        """
        return data.ModelOutput(
            name_model=model_output.name_model,
            recording=model_output.recording,
            detections=self.get_clean_detections(model_output.detections),
        )

    def get_clean_tags(
        self, tags: List[data.PredictedTag]
    ) -> List[data.PredictedTag]:
        """Remove tags with low score."""
        return [
            tag
            for tag in tags
            if tag.confidence_score >= self.detection_threshold
        ]

    def get_clean_detections(
        self, detections: Sequence[data.Detection]
    ) -> List[data.Detection]:
        """Remove detections with low score."""
        return [
            self.clean_detection(detection)
            for detection in detections
            if detection.detection_score >= self.detection_threshold
        ]

    def clean_detection(self, detection: data.Detection) -> data.Detection:
        """Remove tags with low score from detection."""
        return data.Detection(
            id=detection.id,
            prediction_type=detection.prediction_type,
            location=detection.location,
            detection_score=detection.detection_score,
            tags=self.get_clean_tags(detection.tags),
        )



from acoupi.components.audio_recorder.base import TMP_PATH
from acoupi.components.audio_recorder.pipewire_recorder import (
    PWRecorder,
    PWRecorderConfig,
)
from acoupi.components.audio_recorder.pyaudio_recorder import (
    MicrophoneConfig,
    PARecorder,
    PARecorderConfig,
    PyAudioRecorder,
)

__all__ = [
    "PARecorderConfig",
    "PARecorder",
    "MicrophoneConfig",
    "PWRecorderConfig",
    "PWRecorder",
    "PyAudioRecorder",
    "TMP_PATH",
]



"""Recording Saving Filters for acoupi.

RecordingSavingFilters are used to determine if a recording should be saved based on
specific criteria. These filters can be used to save recordings based on time intervals,
detection probabilities, classification probabilities, classification tag values, and more.

The Recording SavingFilters are implemented as classes that inherit from the RecordingSavingFilter
Implementation of the RecordingSavingFilters should implement the should_save_recording method,
which takes a recording object and a list of model outputs, and returns a boolean value.

The RecordingSavingFilters are used in the acoupi.tasks.management module to determine
if a recording should be saved based on the output of the models and the filters provided.
If a recording pass filters, it will be kept and stored in a directory specified by the
RecordingSavingManager. If a recording does not pass the filters, it is deleted.

The RecordingSavingFilters are optional and can be ignored if no recordings should be saved.
"""

import datetime
from typing import List, Optional

from astral import LocationInfo
from astral.sun import sun

from acoupi import data
from acoupi.components import types

__all__ = [
    "SaveIfInInterval",
    "FrequencySchedule",
    "After_DawnDuskTimeInterval",
    "Before_DawnDuskTimeInterval",
    "DetectionTagValue",
    "DetectionTags",
    "SavingThreshold",
]


class SaveIfInInterval(types.RecordingSavingFilter):
    """A time interval RecordingSavingFilter."""

    interval: data.TimeInterval
    """The interval of time where recordings will be saved."""

    timezone: datetime.tzinfo
    """The timezone to use when determining if recording should be saved."""

    def __init__(self, interval: data.TimeInterval, timezone: datetime.tzinfo):
        """Initialise the SaveIfInterval RecordingSavingFilter."""
        self.interval = interval
        self.timezone = timezone

    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        """Save a recording if it falls within the specified interval.

        Examples
        --------
        >>> interval = data.TimeInterval(
        ...     start=datetime.time(21, 30), end=datetime.time(23, 00)
        ... )
        >>> timezone = datetime.timezone.utc
        >>> filter = SaveIfInInterval(interval, timezone)
        >>> recording = data.Recording(
        ...     datetime=datetime.datetime(
        ...         2024, 1, 1, 22, 0, 0, tzinfo=timezone
        ...     )
        ... )
        >>> filter.should_save_recording(recording)
        True
        """
        time = recording.created_on.time()
        if self.interval.start > self.interval.end:
            return self.interval.start <= time or time <= self.interval.end

        return self.interval.start <= time <= self.interval.end


class FrequencySchedule(types.RecordingSavingFilter):
    """A frequency schedule RecordingSavingFilter."""

    duration: int
    """The duration of time (in minutes) where recordings will be saved."""

    frequency: int
    """The frequency of time (in minutes) where recordings will be saved."""

    def __init__(self, duration: int, frequency: int):
        """Initialise the FrequencySchedule RecordingSavingFilter."""
        self.duration = duration
        self.frequency = frequency

    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        """Determine if a recording should be saved."""
        time = recording.created_on
        elapsed_time = (time.minute % self.frequency) + (time.second / 60)
        return elapsed_time < self.duration


class Before_DawnDuskTimeInterval(types.RecordingSavingFilter):
    """A before dawn and dusk time RecordingSavingFilter."""

    def __init__(self, duration: float, timezone: datetime.tzinfo):
        """Initiatlise the Before Dawn & Dusk RecordingSavingFilter."""
        self.duration = duration
        self.timezone = timezone

    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        """Save a recording if it falls within the specified interval, before dawn and dusk.

        Notes
        -----
        The dawn and dusk times are calculated using the `astral` library. The
        `sun` function returns the dawn and dusk times for a specific location, datetime
        and timezone. This information is used to determine the interval before dawn and dusk
        dusk, and whether the current recording falls within this interval.

        Examples
        --------
        >>> DawnTime GMT: 2024-01-01 07:26:00+00:00
        >>> DuskTime GMT: 2024-01-01 16:42:00+00:00
        >>> duration = 30
        >>> timezone = "Europe/London"

        >>> saving_filter = Before_DawnDuskTimeInterval(
        ...     duration, timezone
        ... )
        >>> recording = data.Recording(
        ...     datetime=datetime.datetime(
        ...         2024, 1, 1, 7, 0, 0, tzinfo=timezone
        ...     )
        ... )
        >>> assert saving_filter.should_save_recording(recording)
        True

        >>> saving_filter = Before_DawnDuskTimeInterval(
        ...     duration, timezone
        ... )
        >>> recording = data.Recording(
        ...     datetime=datetime.datetime(
        ...         2024, 1, 1, 17, 0, 0, tzinfo=timezone
        ...     )
        ... )
        >>> assert saving_filter.should_save_recording(recording)
        False
        """
        recording_time = recording.created_on.astimezone(self.timezone)

        sun_info = sun(
            LocationInfo(str(self.timezone)).observer,
            date=recording_time,
            tzinfo=self.timezone,
        )
        dawntime = sun_info["dawn"]
        dusktime = sun_info["dusk"]

        dawntime_interval = dawntime - datetime.timedelta(
            minutes=self.duration
        )
        dusktime_interval = dusktime - datetime.timedelta(
            minutes=self.duration
        )

        return (dawntime_interval <= recording_time <= dawntime) or (
            dusktime_interval <= recording_time <= dusktime
        )


class After_DawnDuskTimeInterval(types.RecordingSavingFilter):
    """An after dawn and dusk time RecordingSavingFilter."""

    duration: float
    """The duration (in minutes) before dawn and dusk where recordings will be saved."""

    timezone: datetime.tzinfo
    """The timezone to use when determining dawntime and dusktime."""

    def __init__(self, duration: float, timezone: datetime.tzinfo):
        """Initiatlise the Before Dawn & Dusk RecordingSavingFilter."""
        self.duration = duration
        self.timezone = timezone

    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        """Save a recording if it falls within the specified interval, after dawn and dusk.

        Notes
        -----
        The dawn and dusk times are calculated using the `astral` library. The
        `sun` function returns the dawn and dusk times for a specific location, datetime
        and timezone. This information is used to determine the interval after dawn and dusk
        dusk, and whether the current recording falls within this interval.

        Examples
        --------
        >>> DawnTime GMT: 2024-01-01 07:26:00+00:00
        >>> DuskTime GMT: 2024-01-01 16:42:00+00:00
        >>> duration = 30
        >>> timezone = "Europe/London"

        >>> saving_filter = After_DawnDuskTimeInterval(
        ...     duration, timezone
        ... )
        >>> recording = data.Recording(
        ...     datetime=datetime.datetime(
        ...         2024, 1, 1, 7, 0, 0, tzinfo=timezone
        ...     )
        ... )
        >>> assert saving_filter.should_save_recording(recording)
        False

        >>> saving_filter = After_DawnDuskTimeInterval(
        ...     duration, timezone
        ... )
        >>> recording = data.Recording(
        ...     datetime=datetime.datetime(
        ...         2024, 1, 1, 17, 0, 0, tzinfo=timezone
        ...     )
        ... )
        >>> assert saving_filter.should_save_recording(recording)
        True
        """
        recording_time = recording.created_on.astimezone(self.timezone)

        sun_info = sun(
            LocationInfo(str(self.timezone)).observer,
            date=recording_time,
            tzinfo=self.timezone,
        )
        dawntime = sun_info["dawn"]
        dusktime = sun_info["dusk"]

        dawntime_interval = dawntime + datetime.timedelta(
            minutes=self.duration
        )
        dusktime_interval = dusktime + datetime.timedelta(
            minutes=self.duration
        )

        return (dawntime <= recording_time <= dawntime_interval) or (
            dusktime <= recording_time <= dusktime_interval
        )


class SavingThreshold(types.RecordingSavingFilter):
    """A SavingThreshold RecordingSavingFilter."""

    saving_threshold: float
    """The score threshold to use."""

    def __init__(self, saving_threshold: float):
        """Initialise the RecordingSavingFilter."""
        self.saving_threshold = saving_threshold

    def has_confident_model_output(
        self, model_output: data.ModelOutput
    ) -> bool:
        """Determine if a model output has confident detections.

        An output is considered confident if any detection score is greater
        than or equal to the threshold.

        Parameters
        ----------
        model_output : data.ModelOutput
            The model output of the recording containing detections and tags.

        Returns
        -------
        bool
            True if any detection score is above the saving threshold.
            False if no detection score is above the saving threshold.
        """
        if any(
            detection.detection_score >= self.saving_threshold
            for detection in model_output.detections
        ):
            return True

        return False

    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        """Save a recording if it contains any confident detections."""
        if model_outputs is None:
            return False

        return any(
            self.has_confident_model_output(model_output)
            for model_output in model_outputs
        )


class DetectionTagValue(types.RecordingSavingFilter):
    """A RecordingSavingFilter that keeps recordings with specific tag values."""

    values: List[str]
    """The tag values to focus on."""

    def __init__(self, values: List[str]):
        """Initialise the RecordingSavingFilter."""
        self.values = values

    def has_confident_tagvalues(self, model_output: data.ModelOutput) -> bool:
        """Determine if a model output has a confident tag values.

        An output is considered confident if any of its tag value (e.g., species_name)
        is in the values list.

        Parameters
        ----------
        model_output : data.ModelOutput
            The model output of the recording containing detections and tags.

        Returns
        -------
            bool
        """
        for detection in model_output.detections:
            for tag in detection.tags:
                if tag.tag.value in self.values:
                    return True
        return False

    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        """Save a recording if it contains any confident tag values."""
        if model_outputs is None:
            return False

        return any(
            self.has_confident_tagvalues(model_output)
            for model_output in model_outputs
        )


class DetectionTags(types.RecordingSavingFilter):
    """A RecordingSavingFilter that keeps recordings with selected tags.

    This filter will keep recordings that contain confident tag
    predictions that are in the tags list.
    """

    tags: List[data.Tag]
    """The tags to focus on."""

    saving_threshold: float
    """The score threshold to use."""

    def __init__(self, tags: List[data.Tag], saving_threshold: float = 0.5):
        """Initialise the RecordingSavingFilter."""
        self.tags = tags
        self.saving_threshold = saving_threshold

    def has_confident_tag(self, model_output: data.ModelOutput) -> bool:
        """Determine if a model output has a confident tag.

        An output is considered confident if any of its detections have a
        matching tag with score greater than or equal to the threshold.

        Parameters
        ----------
        model_output : data.ModelOutput
            The model output of the recording containing detections and tags.

        Returns
        -------
            bool
        """
        for detection in model_output.detections:
            if detection.detection_score < self.saving_threshold:
                continue

            for tag in detection.tags:
                if tag.tag not in self.tags:
                    continue

                if tag.confidence_score >= self.saving_threshold:
                    return True

        return False

    def should_save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> bool:
        """Save a recording if it contains any confident tags or detections."""
        if model_outputs is None:
            return False

        return any(
            self.has_confident_tag(model_output)
            for model_output in model_outputs
        )



"""Storages for acoupi.

Storages are used to store recordings and detections locally. The
stores keep track of the recordings and detections that have been
made. The stored data can be retrieved later to be sent to a
remote server.

Storages are implemented as classes that inherit from Storage.
The class should implement the methods for storing and retrieving
data, as well as the methods for retrieving the current deployment
and the recordings and detections for a given deployment. See
the Storage class for more details.

"""

from acoupi.components.stores.sqlite import SqliteStore

__all__ = ["SqliteStore"]



"""System datetime management."""

import subprocess
from typing import Optional
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

__all__ = ["get_system_timezone", "set_system_timezone"]


def get_system_timezone() -> Optional[ZoneInfo]:
    """Get the current system timezone.

    Returns
    -------
    Optional[str]
        The timezone string (e.g., 'Europe/London') or None if it cannot be
        determined.
    """
    try:
        result = subprocess.run(
            ["timedatectl", "show", "--property=Timezone", "--value"],
            capture_output=True,
            text=True,
            check=True,
        )
        tz_string = result.stdout.strip()

    except subprocess.CalledProcessError:
        return None

    if not tz_string:
        return None

    try:
        return ZoneInfo(tz_string)

    except ZoneInfoNotFoundError:
        return None


def set_system_timezone(timezone_str: str) -> bool:
    """Set the system timezone using timedatectl.

    Parameters
    ----------
    timezone_str : str
        The timezone to set (e.g., 'America/New_York').

    Returns
    -------
    bool
        True if successful, False otherwise.
    """
    try:
        ZoneInfo(timezone_str)
    except ZoneInfoNotFoundError:
        return False

    try:
        subprocess.run(
            ["sudo", "timedatectl", "set-timezone", timezone_str],
            check=True,
            capture_output=True,
        )
        return True
    except subprocess.CalledProcessError:
        return False



import datetime
from collections import namedtuple
from unittest.mock import Mock

from acoupi import data
from acoupi.devices import metrics


class TestGetRemainingStorage:
    def test_returns_metric(self, monkeypatch, patched_now):
        """Return a storage metric containing free bytes."""
        now = datetime.datetime(2024, 1, 1, tzinfo=datetime.timezone.utc)
        patched_now(now)
        usage = namedtuple("usage", ["total", "used", "free"])(100, 40, 60)
        monkeypatch.setattr(metrics.shutil, "disk_usage", lambda path: usage)

        metric = metrics.get_remaining_storage(path=Mock())

        assert metric == data.Metric(
            name="remaining_storage",
            value=60,
            unit="bytes",
            captured_on=now,
        )


class TestGetUsedStorage:
    def test_returns_metric(self, monkeypatch, patched_now):
        """Return a storage metric containing used bytes."""
        now = datetime.datetime(2024, 1, 1, tzinfo=datetime.timezone.utc)
        patched_now(now)
        usage = namedtuple("usage", ["total", "used", "free"])(100, 40, 60)
        monkeypatch.setattr(metrics.shutil, "disk_usage", lambda path: usage)

        metric = metrics.get_used_storage(path=Mock())

        assert metric == data.Metric(
            name="used_storage",
            value=40,
            unit="bytes",
            captured_on=now,
        )


class TestGetFreeMemory:
    def test_returns_metric(self, monkeypatch, patched_now):
        """Return a memory metric containing available bytes."""
        now = datetime.datetime(2024, 1, 1, tzinfo=datetime.timezone.utc)
        patched_now(now)
        memory = Mock(available=128)
        monkeypatch.setattr(metrics.psutil, "virtual_memory", lambda: memory)

        metric = metrics.get_free_memory()

        assert metric == data.Metric(
            name="free_memory",
            value=128,
            unit="bytes",
            captured_on=now,
        )


class TestConsumedMemory:
    def test_returns_metric(self, monkeypatch, patched_now):
        """Return a memory metric containing process RSS bytes."""
        now = datetime.datetime(2024, 1, 1, tzinfo=datetime.timezone.utc)
        patched_now(now)
        process = Mock()
        process.memory_info.return_value = Mock(rss=256)
        monkeypatch.setattr(metrics.os, "getpid", lambda: 123)
        monkeypatch.setattr(metrics.psutil, "Process", lambda pid: process)

        metric = metrics.consumed_memory()

        assert metric == data.Metric(
            name="consumed_memory",
            value=256,
            unit="bytes",
            captured_on=now,
        )


class TestGetCpuUsage:
    def test_returns_metric(self, monkeypatch, patched_now):
        """Return a CPU metric containing process usage percent."""
        now = datetime.datetime(2024, 1, 1, tzinfo=datetime.timezone.utc)
        patched_now(now)
        process = Mock()
        process.cpu_percent.return_value = 12.5
        monkeypatch.setattr(metrics.os, "getpid", lambda: 123)
        monkeypatch.setattr(metrics.psutil, "Process", lambda pid: process)

        metric = metrics.get_cpu_usage()

        assert metric == data.Metric(
            name="cpu_usage",
            value=12.5,
            unit="percent",
            captured_on=now,
        )



"""Acoupi Default Program.

This is the most basic acoupi program. It only records audio files and does not
do any processing and messaging.
"""

import datetime
from typing import Optional
from zoneinfo import ZoneInfo

from pydantic import BaseModel

from acoupi import components, data
from acoupi.programs.templates import BasicProgram, BasicProgramConfiguration


class SaveRecordingFilter(BaseModel):
    """Recording saving options configuration."""

    starttime: datetime.time = datetime.time(hour=18, minute=0, second=0)

    endtime: datetime.time = datetime.time(hour=20, minute=0, second=0)

    before_dawndusk_duration: int = 10

    after_dawndusk_duration: int = 10

    frequency_duration: int = 5

    frequency_interval: int = 30


class ConfigSchema(BasicProgramConfiguration):
    """Configuration Schema for Test Program."""

    recording_saving: Optional[SaveRecordingFilter] = None


class Program(BasicProgram):
    """Test Program."""

    config_schema = ConfigSchema

    ### --- Configure Additional Filters - SavingRecordingFilters --- ###
    def get_recording_filters(self, config: ConfigSchema):
        if not config.recording_saving:
            # No saving filters defined
            return []

        saving_filters = []
        timezone = ZoneInfo(config.timezone)
        recording_saving = config.recording_saving

        # Main filter
        # Will only save recordings if the recording time is in the
        # interval defined by the start and end time.
        if (
            recording_saving.starttime is not None
            and recording_saving.endtime is not None
        ):
            saving_filters.append(
                components.SaveIfInInterval(
                    interval=data.TimeInterval(
                        start=recording_saving.starttime,
                        end=recording_saving.endtime,
                    ),
                    timezone=timezone,
                )
            )

        # Additional filters
        if (
            recording_saving.frequency_duration != 0
            and recording_saving.frequency_interval != 0
        ):
            # This filter will only save recordings at a frequency defined
            # by the duration and interval.
            saving_filters.append(
                components.FrequencySchedule(
                    duration=recording_saving.frequency_duration,
                    frequency=recording_saving.frequency_interval,
                )
            )

        if recording_saving.before_dawndusk_duration != 0:
            # This filter will only save recordings if the recording time
            # is before dawn or dusk.
            saving_filters.append(
                components.Before_DawnDuskTimeInterval(
                    duration=recording_saving.before_dawndusk_duration,
                    timezone=timezone,
                )
            )

        if recording_saving.after_dawndusk_duration != 0:
            # This filter will only save recordings if the recording time
            # is after dawn or dusk.
            saving_filters.append(
                components.After_DawnDuskTimeInterval(
                    duration=recording_saving.after_dawndusk_duration,
                    timezone=timezone,
                )
            )

        return saving_filters



"""Sqlite storage backend for acoupi.

The SqliteStore class is used to store the recordings, detections and
deployments locally. The data is stored in a sqlite database file in the given
path.

Whenever a recording or a detection is made, it should be stored to
keep track of the data. Current deployment information can be retrieved and
updated.

To use the store, create an instance of the SqliteStore class, and use the
methods to store and retrieve data.

Example:
    Load the store:

    >>> from acoupi.storages.sqlite import SqliteStore

    Create a new store instance, pointing to a database file:

    >>> store = SqliteStore("acoupi.db")

    Get the current deployment:

    >>> deployment = store.get_current_deployment()
    >>> deployment
    Deployment(id=1, started_on=datetime.datetime(2021, 1, 1, 0, 0), latitude=0, longitude=0)

    Create a new deployment:

    >>> deployment = Deployment(
    ...     started_on=datetime.datetime(2021, 1, 1, 0, 0),
    ...     device_id="0000000000000000",
    ...     latitude=0.0,
    ...     longitude=0.0,
    ... )
    >>> store.store_deployment(deployment)

    Store a new recording:

    >>> recording = Recording(
    ...     datetime=datetime.datetime(2021, 1, 1, 0, 0),
    ...     duration=10,
    ...     samplerate=44100,
    ...     audio_channels=1,
    ... )
    >>> store.store_recording(recording)

    Store a new detection:

    >>> detection = Detection(
    ...     species_name="species",
    ...     score=0.5,
    ... )
    >>> store.store_detections(recording, [detection])

"""

from acoupi.components.stores.sqlite.store import SqliteStore

__all__ = [
    "SqliteStore",
]



"""Messengers for acoupi.

Messengers are responsible for sending messages to external services. The
messengers are templates illustrating how to send messages using different
communication protocols (e.g., MQTT, HTTP).

The messengers are implemented as classes that inherit from Messenger. The
class should implement the send_message method, which takes a message and sends
it to the external service. The class should also implement the check method,
which checks the connection status of the messenger.

The MQTTMessenger sends messages using the MQTT protocol. The HTTPMessenger
sends messages using HTTP POST requests.
"""

import json
import logging
import socket
from enum import Enum
from typing import Literal, Optional

import paho.mqtt.client as mqtt
import requests
from celery.utils.log import get_task_logger
from paho.mqtt.enums import CallbackAPIVersion, MQTTErrorCode
from pydantic import BaseModel, SecretStr, field_serializer

from acoupi import data
from acoupi.components import types
from acoupi.devices import get_device_id
from acoupi.system.exceptions import HealthCheckError, MessageSendError

__all__ = [
    "MQTTMessenger",
    "HTTPMessenger",
]


class MQTTTransport(Enum):
    TCP = "tcp"
    WEBSOCKETS = "websockets"
    UNIX = "unix"


class MQTTConfig(BaseModel):
    host: str
    username: str
    password: Optional[SecretStr] = None
    topic: str = "acoupi"
    port: int = 1884
    timeout: int = 5
    use_tls: bool = False
    transport: MQTTTransport = MQTTTransport.TCP

    @field_serializer("password", when_used="json")
    def dump_password(self, value):
        return value.get_secret_value() if value else None


class MQTTMessenger(types.Messenger):
    """Messenger that sends messages via MQTT."""

    client: mqtt.Client
    """The MQTT client."""

    topic: str
    """The MQTT topic to send messages to."""

    timeout: int
    """Timeout for sending messages."""

    logger: logging.Logger

    def __init__(
        self,
        host: str,
        topic: str,
        port: int = 1884,
        username: Optional[str] = None,
        password: Optional[str] = None,
        timeout: int = 5,
        use_tls: bool = False,
        logger: Optional[logging.Logger] = None,
        transport: Literal["tcp", "websockets", "unix"] = "tcp",
    ) -> None:
        """Initialise the MQTT messenger.

        Parameters
        ----------
        host : str
            The host to connect to. Example: "mqtt.localhost.org".
        username : str
            The username to authenticate with.
        topic : str
            The topic to send messages to. Example: "org/survey/device_00/".
        port : int, optional
            The port to connect to, by default 1884.
        password : Optional[SecretStr], optional
            The password to authenticate with, by default None.
        use_tls: bool
            Use TLS is host requires this with local certs (eg. HiveHQ) - default sets to false

        Notes
        -----
        Will use the device ID as the client ID.
        """
        self.topic = topic
        self.timeout = timeout
        self.host = host
        self.port = port
        self.client_id = get_device_id()
        self.use_tls = use_tls
        self.transport = transport

        self.client = mqtt.Client(
            callback_api_version=CallbackAPIVersion.VERSION2,
            client_id=self.client_id,
            clean_session=False,
            transport=self.transport,
        )

        self.client.username_pw_set(username, password)

        if logger is None:
            logger = get_task_logger(__name__)

        self.logger = logger

        if self.use_tls:
            # Calling tls_set() without arguments uses the system's
            # default CA certificates, which works for HiveHQ
            self.logger.info("Using TLS Connection.")
            self.client.tls_set()

    @classmethod
    def from_config(
        cls,
        config: MQTTConfig,
        logger: Optional[logging.Logger] = None,
    ):
        """Create an MQTTMessenger from a configuration object."""
        return cls(
            host=config.host,
            port=config.port,
            username=config.username,
            password=config.password.get_secret_value()
            if config.password
            else None,
            topic=config.topic,
            timeout=config.timeout,
            use_tls=config.use_tls,
            transport=config.transport.value,
            logger=logger,
        )

    def check_connection(self) -> MQTTErrorCode:
        """Check the connection status of the MQTT client."""
        if self.client.is_connected():
            return MQTTErrorCode.MQTT_ERR_SUCCESS

        # NOTE: According to the docs, the host attribute is only set after
        # the connect method is called. If the host attribute is set, then
        # its better to use the reconnect method.
        if self.client.host == self.host:
            return self.client.reconnect()

        return self.client.connect(self.host, port=self.port)

    def send_message(self, message: data.Message) -> data.Response:
        """Send a recording message.

        Parameters
        ----------
        message : data.Message
            The message to send.

        Returns
        -------
        data.Response
            A response containing the message, status, content, and received
            time.

        Examples
        --------
        >>> message = data.Message(
        ...     content="hello world",
        ... )
        >>> messenger = MQTTMessenger(
        ...     host="mqtt.localhost.org",
        ...     username="mqttusername",
        ...     topic="org/survey/device_00",
        ...     clientid="org/survey/device_00",
        ... )
        >>> messenger.send_message(message)
        >>> data.Response(
        ...     message=data.Message(content="{}"),
        ...     status=ResponseStatus.SUCCESS,
        ...     content="MQTT_ERR_SUCCESS",
        ...     received_on=datetime.datetime(),
        ... )
        """
        mqtt_status = self.check_connection()

        if mqtt_status != MQTTErrorCode.MQTT_ERR_SUCCESS:
            raise MessageSendError(
                f"MQTT connection error: {MQTTErrorCode(mqtt_status).name}"
            )

        response = self.client.publish(
            topic=self.topic,
            payload=message.content,
        )

        status = data.ResponseStatus.SUCCESS
        try:
            response.wait_for_publish(timeout=5)
        except ValueError as e:
            raise MessageSendError(
                f"MQTT publish failed for message {message.id}: {e}"
            ) from e
        except RuntimeError as e:
            raise MessageSendError(
                f"MQTT publish failed for message {message.id}: {e}"
            ) from e

        if response.rc != MQTTErrorCode.MQTT_ERR_SUCCESS:
            status = data.ResponseStatus.ERROR

        received_on = data.utc_now()
        logging.debug(f"Message sent: {message.content}")

        return data.Response(
            message=message,
            status=status,
            content=MQTTErrorCode(response.rc).name,
            received_on=received_on,
        )

    def check(self) -> None:
        """Check the connection status of the MQTT client.

        Raises
        ------
        HealthCheckError
            If the connection is not successful. This could be due to a
            connection error or an authentication failure.
        """
        try:
            mqtt_status = self.check_connection()
        except socket.gaierror as err:
            raise HealthCheckError(
                "Health check failed: Unable to resolve the MQTT broker host."
            ) from err
        except socket.timeout as err:
            raise HealthCheckError(
                "Health check failed: Timeout connecting to the MQTT broker."
            ) from err

        if mqtt_status != MQTTErrorCode.MQTT_ERR_SUCCESS:
            error_name = MQTTErrorCode(mqtt_status).name
            raise HealthCheckError(
                f"Health check failed: MQTT Connection Error ({error_name}).\n"
                "Possible causes:\n"
                "  * Broker unavailable: Ensure the MQTT broker is running "
                "and accessible.\n"
                "  * Authentication failed: Verify your MQTT credentials "
                "(username/password) are correct.\n"
                "  * Network issues: Check your network connection and any "
                "firewall settings.\n"
                "  * Misconfiguration: Review your MQTT configuration "
                "parameters for accuracy."
            )

        # # Wait for the connection to be established or disconnected due to
        # # failed authentication
        # time.sleep(0.1)
        #
        # if not self.client.is_connected():
        #     raise HealthCheckError(
        #         "Health check failed: MQTT Connection Error (Authentication "
        #         "Failed).\n"
        #         "The MQTT client was unable to authenticate with the broker. "
        #         "Verify your MQTT credentials (username/password) are correct."
        #     )


class HTTPConfig(BaseModel):
    base_url: str
    content_type: str = "application/json"
    timeout: int = 5


class HTTPMessenger(types.Messenger):
    """Messenger that sends messages via HTTP POST requests."""

    base_url: str
    """The base URL to send messages to."""

    timeout: int
    """Timeout for sending messages in seconds."""

    base_params: dict
    """Base parameters to send with each request."""

    headers: dict
    """Headers to send with each request."""

    def __init__(
        self,
        base_url: str,
        base_params: Optional[dict] = None,
        headers: Optional[dict] = None,
        timeout: int = 5,
        content_type: str = "application/json",
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """Initialise the HTTP messenger.

        Parameters
        ----------
        base_url : str
            The URL to send messages to. This should include the protocol
            (e.g. http:// or https://) and the hostname (e.g. localhost)
            and the path (e.g. /api/endpoint).
        base_params : Optional[dict], optional
            Base parameters to send with each request, by default None.
        headers : Optional[dict], optional
            Headers to send with each request, by default None.
        timeout : int, optional
            Seconds to wait for a response before timing out, by default 5.
        content_type : str, optional
            The content type to send with each request, by default
            "application/json".

        """
        self.base_url = base_url
        self.timeout = timeout
        self.base_params = base_params or {}
        self.headers = headers or {}

        # Set the content type if not already set
        if self.headers.get("Content-Type") is None:
            self.headers["Content-Type"] = content_type

        # Set accepted content type if not already set
        if self.headers.get("Accept") is None:
            self.headers["Accept"] = content_type

        self.content_type = self.headers["Content-Type"]

        if logger is None:
            logger = get_task_logger(__name__)

        self.logger = logger

    @classmethod
    def from_config(
        cls,
        config: HTTPConfig,
        logger: Optional[logging.Logger] = None,
    ):
        """Create an HTTPMessenger from a configuration object."""
        return cls(
            base_url=config.base_url,
            timeout=config.timeout,
            content_type=config.content_type,
            logger=logger,
        )

    def send_message(self, message: data.Message) -> data.Response:
        """Send a recording message through a HTTP POST request."""
        status = data.ResponseStatus.SUCCESS
        response_content = None

        try:
            message_content = message.content
            if isinstance(message_content, bytes):
                message_content = message_content.decode("utf-8")

            content_json = json.loads(message_content)

            if self.content_type == "application/json":
                response = requests.post(
                    self.base_url,
                    json=content_json,
                    params=self.base_params,
                    headers=self.headers,
                    timeout=self.timeout,
                )

            else:
                response = requests.post(
                    self.base_url,
                    data=content_json,
                    params=self.base_params,
                    headers=self.headers,
                    timeout=self.timeout,
                )

            response_content = response.text

            if not response.ok:
                status = data.ResponseStatus.ERROR

        except (UnicodeDecodeError, ValueError) as error:
            raise MessageSendError(
                f"Invalid HTTP message payload for message {message.id}: {error}"
            ) from error
        except requests.exceptions.RequestException as error:
            raise MessageSendError(
                f"HTTP request failed for message {message.id}: {error}"
            ) from error
        except RuntimeError as error:
            raise MessageSendError(
                f"HTTP send failed for message {message.id}: {error}"
            ) from error

        received_on = data.utc_now()

        return data.Response(
            content=response_content,
            message=message,
            status=status,
            received_on=received_on,
        )

    def check(self) -> None:
        """Check the connection status of the HTTP client.

        Raises
        ------
        HealthCheckError
            If the connection is not successful. This could be due to a
            connection error or if the POST method is not allowed.
        """
        try:
            response = requests.options(self.base_url)
        except requests.exceptions.ConnectionError as err:
            raise HealthCheckError(
                f"Health check failed: Unable to connect to {self.base_url}.\n"
                "Possible causes:\n"
                f"  * Incorrect URL: Double-check that '{self.base_url}' is "
                "the correct address.\n"
                "  * Network issues: Verify your internet connection or any "
                "firewall settings.\n"
                "  * Service down: The server at {self.base_url} may be "
                "temporarily unavailable."
            ) from err

        if not response.ok:
            status_code = response.status_code
            error_type = "Client" if 400 <= status_code < 500 else "Server"
            raise HealthCheckError(
                f"Health check failed for {self.base_url}: {error_type} "
                f"Error ({status_code}).\n"
                "Could connect to the server but received an error response.\n"
                "Possible causes:\n"
                "  * 4xx Client Error: Missing authentication credentials, or "
                "authorization issues.\n"
                f"  * 5xx Server Error:  The service at {self.base_url} is "
                "temporarily unavailable or experiencing issues."
            )

        if (
            "Allow" in response.headers
            and "POST" not in response.headers["Allow"]
        ):
            raise HealthCheckError(
                f"Could connect to {self.base_url} but POST method is "
                "not allowed. Check the server configuration."
            )



from acoupi.system.constants import CeleryConfig


def test_celery_config_defaults_include_short_beat_loop_interval():
    config = CeleryConfig().model_dump()

    assert config["beat_max_loop_interval"] == 1



import datetime

import pytest

from acoupi.tasks.schedules import aligned_schedule

UTC = datetime.timezone.utc


def test_aligned_schedule_is_due_on_interval_boundary():
    schedule = aligned_schedule(
        run_every=datetime.timedelta(seconds=10),
        nowfun=lambda: datetime.datetime(2024, 1, 1, 12, 0, 20, tzinfo=UTC),
    )

    state = schedule.is_due(
        last_run_at=datetime.datetime(2024, 1, 1, 12, 0, 19, tzinfo=UTC)
    )

    assert state.is_due is True
    assert state.next == 10


def test_aligned_schedule_waits_until_next_boundary():
    schedule = aligned_schedule(
        run_every=datetime.timedelta(seconds=10),
        nowfun=lambda: datetime.datetime(2024, 1, 1, 12, 0, 12, tzinfo=UTC),
    )

    state = schedule.is_due(
        last_run_at=datetime.datetime(2024, 1, 1, 12, 0, 10, tzinfo=UTC)
    )

    assert state.is_due is False
    assert state.next == pytest.approx(8)


def test_aligned_schedule_honours_offset_seconds():
    schedule = aligned_schedule(
        run_every=datetime.timedelta(seconds=10),
        offset_seconds=5,
        nowfun=lambda: datetime.datetime(2024, 1, 1, 12, 0, 25, tzinfo=UTC),
    )

    state = schedule.is_due(
        last_run_at=datetime.datetime(2024, 1, 1, 12, 0, 24, tzinfo=UTC)
    )

    assert state.is_due is True
    assert state.next == 10


def test_aligned_schedule_does_not_repeat_in_same_slot():
    schedule = aligned_schedule(
        run_every=datetime.timedelta(seconds=10),
        nowfun=lambda: datetime.datetime(
            2024, 1, 1, 12, 0, 20, 500000, tzinfo=UTC
        ),
    )

    state = schedule.is_due(
        last_run_at=datetime.datetime(2024, 1, 1, 12, 0, 20, tzinfo=UTC)
    )

    assert state.is_due is False
    assert state.next == pytest.approx(9.5)


def test_aligned_schedule_rejects_invalid_offset():
    with pytest.raises(
        ValueError, match="offset_seconds must be smaller than run_every"
    ):
        aligned_schedule(
            run_every=datetime.timedelta(seconds=10),
            offset_seconds=10,
        )



from acoupi.system.config.operations import (
    dump_config,
    get_config_field,
    load_config,
    set_config_field,
    write_config,
)
from acoupi.system.config.parsers import parse_config_from_args

__all__ = [
    "dump_config",
    "get_config_field",
    "load_config",
    "parse_config_from_args",
    "set_config_field",
    "write_config",
]



"""Schema helpers for the SQLite message store."""

import sqlite3


def create_message_schema(connection: sqlite3.Connection) -> None:
    """Create the message-store schema if it does not exist."""
    connection.executescript(
        """
        CREATE TABLE IF NOT EXISTS message (
            id BLOB PRIMARY KEY,
            content BLOB NOT NULL,
            created_on TEXT NOT NULL
        );

        CREATE TABLE IF NOT EXISTS response (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            content TEXT,
            message_id BLOB NOT NULL,
            status INTEGER NOT NULL,
            received_on TEXT NOT NULL,
            FOREIGN KEY (message_id) REFERENCES message(id)
        );

        CREATE INDEX IF NOT EXISTS idx_response_message_id
        ON response(message_id);

        CREATE INDEX IF NOT EXISTS idx_response_status
        ON response(status);
        """
    )



import pytest

from acoupi.system import Settings, exceptions, lifecycle, programs


def test_setup_program_creates_all_the_required_files(
    settings: Settings,
):
    lifecycle.setup_program(
        settings,
        program_name="acoupi.programs.test",
        prompt=False,
    )

    assert settings.home.exists()
    assert settings.program_file.exists()

    # The program name is stored in a file
    assert settings.program_name_file.exists()
    assert settings.program_name_file.read_text() == "acoupi.programs.test"

    # Check that configurations are created
    assert settings.program_config_file.exists()
    assert settings.celery_config_file.exists()

    # Check that scripts are created
    assert settings.beat_script_path.exists()
    assert settings.start_script_path.exists()
    assert settings.stop_script_path.exists()
    assert settings.restart_script_path.exists()


def test_start_program_cleans_deployment_if_failure_before_start(
    settings: Settings,
    monkeypatch,
):
    def mock_load_program(_: Settings):
        raise exceptions.InvalidProgramError(program="test")

    monkeypatch.setattr(
        programs,
        "load_program",
        mock_load_program,
    )

    with pytest.raises(exceptions.InvalidProgramError):
        lifecycle.start_program(settings, name="test")

    assert not settings.deployment_file.exists()



"""Test acoupi components saving filters."""

from pathlib import Path
from typing import List

import pytest

from acoupi import data
from acoupi.components import message_factories


@pytest.fixture
def create_test_detection():
    """Fixture for creating random detections.

    Will create detections with no location and a single tag.
    """

    def factory(
        detection_score: float = 0.8,
        tag_key: str = "species",
        tag_value: str = "Myotis myotis",
        confidence_score: float = 0.6,
    ) -> data.Detection:
        """Return a random detection."""
        return data.PresenceDetection(
            location=data.BoundingBox(
                coordinates=(0.1, 0.2, 0.3, 0.4),
            ),
            detection_score=detection_score,
            tags=[
                data.PredictedTag(
                    tag=data.Tag(
                        key=tag_key,
                        value=tag_value,
                    ),
                    confidence_score=confidence_score,
                )
            ],
        )

    return factory


@pytest.fixture
def create_test_model_output():
    """Fixture for creating random model output."""
    deployment = data.Deployment(
        name="test_deployment",
    )

    recording = data.Recording(
        path=Path("test.wav"),
        duration=1,
        samplerate=256000,
        audio_channels=1,
        deployment=deployment,
        created_on=data.utc_now(),
    )

    def factory(
        detections: List[data.Detection],
    ) -> data.ModelOutput:
        """Return a model output."""
        return data.ModelOutput(
            recording=recording,
            name_model="test_model",
            detections=detections,
        )

    return factory


"""TESTS - DETECTION THRESHOLD - MESSAGE BUILDERS"""


def test_message_builder_detections_below_threshold(
    create_test_detection,
    create_test_model_output,
):
    """Test that detections below the threshold are removed from the message."""
    detection_threshold = 0.6
    message_builder = message_factories.DetectionThresholdMessageBuilder(
        detection_threshold=detection_threshold,
    )

    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                detection_score=0.5,
                tag_key="species",
                tag_value="species_1",
                confidence_score=0.4,
            ),
        ]
    )

    message = message_builder.build_message(model_output)

    assert message is None


def test_message_builder_detections_with_mixthreshold(
    create_test_detection,
    create_test_model_output,
):
    """Test that detections below the threshold are removed from the message."""
    detection_threshold = 0.6
    message_builder = message_factories.DetectionThresholdMessageBuilder(
        detection_threshold=detection_threshold,
    )

    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                detection_score=0.5,
                tag_key="species",
                tag_value="species_1",
                confidence_score=0.4,
            ),
            create_test_detection(
                detection_score=0.7,
                tag_key="species",
                tag_value="species_2",
                confidence_score=0.6,
            ),
        ]
    )
    clean_detections = message_builder.filter_detections(
        model_output.detections
    )
    assert len(clean_detections) == 1

    message = message_builder.build_message(model_output)
    assert message is not None



"""Acoupi program templates.

This module provides base classes and configuration schemas to simplify the
creation of Acoupi programs.

Available Templates:

- **Basic Program:** Provides a foundation for building Acoupi programs,
  including features for audio recording, metadata storage, and file management.
  - Base class: `BasicProgram`
  - Configuration schema: `BasicProgramConfiguration`
- **Messaging Program:** Extends the `BasicProgram` with messaging
  capabilities, enabling programs to send messages and heartbeats via HTTP or
  MQTT.
  - Base class: `MessagingProgram`
  - Configuration schema: `MessagingProgramConfiguration`
- **Detection Program:**  Extends the `MessagingProgram` with audio detection
  capabilities, allowing programs to run detection models on recordings and
  generate messages based on the results.
  - Base class: `DetectionProgram`
  - Configuration schema: `DetectionProgramConfiguration`

Each template includes a base class that provides core functionality and a
configuration schema to define the program's settings.

For detailed usage instructions, customization options, and examples, refer
to the individual template documentation.
"""

from acoupi.programs.templates.basic import (
    AudioConfiguration,
    BasicProgram,
    BasicProgramConfiguration,
    PathsConfiguration,
)
from acoupi.programs.templates.detection import (
    DetectionProgram,
    DetectionProgramConfiguration,
    DetectionsConfiguration,
)
from acoupi.programs.templates.messaging import (
    MessagingConfig,
    MessagingProgram,
    MessagingProgramConfiguration,
)

__all__ = [
    "AudioConfiguration",
    "BasicProgram",
    "BasicProgram",
    "BasicProgramConfiguration",
    "DetectionProgram",
    "DetectionProgramConfiguration",
    "DetectionsConfiguration",
    "MessagingConfig",
    "MessagingProgram",
    "MessagingProgramConfiguration",
    "PathsConfiguration",
]



"""Test the recording of an audio files."""

import math
import wave
from pathlib import Path

import guano
import pytest

from acoupi import components, data
from acoupi.devices.audio.pyaudio import (
    get_default_microphone,
    has_input_audio_device,
)
from acoupi.system.exceptions import HealthCheckError
from acoupi.tasks.recording import add_guano_metadata


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
def test_audio_recording(deployment: data.Deployment, tmp_path: Path):
    """Test getting information from microhpone."""
    audio_channels, samplerate, device_name = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=0.1,
        samplerate=samplerate,
        audio_channels=audio_channels,
        device_name=device_name,
        chunksize=8192,
        audio_dir=tmp_path,
    )

    # record audio
    recording = recorder.record(deployment=deployment)
    assert isinstance(recording, data.Recording)
    assert recording.deployment == deployment
    assert recording.duration == 0.1
    assert recording.samplerate == samplerate
    assert recording.audio_channels == audio_channels
    assert recording.path is not None
    assert recording.path.exists()


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
def test_check_is_succesful(tmp_path: Path):
    """Test check_is_succesful."""
    audio_channels, samplerate, device_name = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=0.1,
        samplerate=samplerate,
        audio_channels=audio_channels,
        device_name=device_name,
        chunksize=8192,
        audio_dir=tmp_path,
    )

    # If everything is ok, the check should pass
    recorder.check()


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
def test_check_fails_if_recording_duration_is_zero(
    monkeypatch, tmp_path: Path
):
    audio_channels, samplerate, device_name = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=1,
        samplerate=samplerate,
        audio_channels=audio_channels,
        device_name=device_name,
        chunksize=8192,
        audio_dir=tmp_path,
    )

    def mock_record_audio(*args, **kwargs) -> bytes:
        return b""

    monkeypatch.setattr(
        "acoupi.components.audio_recorder.pyaudio_recorder.record_audio",
        mock_record_audio,
    )

    # If the duration is zero, the check should fail
    with pytest.raises(HealthCheckError):
        recorder.check()


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
def test_check_fails_if_invalid_samplerate(tmp_path: Path):
    audio_channels, _, device_name = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=1,
        samplerate=10,
        audio_channels=audio_channels,
        device_name=device_name,
        chunksize=8192,
        audio_dir=tmp_path,
    )

    # If the samplerate is invalid, the check should fail
    with pytest.raises(HealthCheckError, match="selected samplerate"):
        recorder.check()


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
def test_check_fails_if_audio_device_is_not_found(tmp_path: Path):
    audio_channels, samplerate, _ = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=1,
        samplerate=samplerate,
        audio_channels=audio_channels,
        device_name="invalid_device",
        chunksize=8192,
        audio_dir=tmp_path,
    )

    with pytest.raises(HealthCheckError, match="device"):
        recorder.check()


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
def test_check_fails_if_invalid_number_of_audio_channels(tmp_path: Path):
    _, samplerate, device_name = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=1,
        samplerate=samplerate,
        audio_channels=1000,
        device_name=device_name,
        chunksize=8192,
        audio_dir=tmp_path,
    )

    with pytest.raises(HealthCheckError, match="channels"):
        recorder.check()


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
@pytest.mark.parametrize("duration", [0.1, 0.3, 0.7, 1])
def test_recording_duration_is_as_requested(
    tmp_path: Path,
    duration: float,
    deployment: data.Deployment,
):
    _, samplerate, device_name = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=duration,
        samplerate=samplerate,
        audio_channels=1,
        device_name=device_name,
        chunksize=8192,
        audio_dir=tmp_path,
    )

    recording = recorder.record(deployment=deployment)

    assert recording.path is not None
    with wave.open(str(recording.path)) as wf:
        assert samplerate == wf.getframerate()
        assert math.isclose(
            duration,
            wf.getnframes() / samplerate,
            abs_tol=0.01,
        )


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
def test_can_record_with_time_expansion(
    tmp_path: Path, deployment: data.Deployment
):
    _, samplerate, device_name = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=1,
        samplerate=samplerate,
        audio_channels=1,
        device_name=device_name,
        chunksize=8192,
        audio_dir=tmp_path,
        time_expansion=0.1,
    )

    recording = recorder.record(deployment=deployment)

    # The recording object has the original samplerate
    assert recording.samplerate == samplerate
    assert recording.time_expansion == 0.1

    with wave.open(str(recording.path)) as wf:
        stored_samplerate = wf.getframerate()
        assert math.isclose(
            stored_samplerate,
            recorder.get_expanded_samplerate(),
            abs_tol=0.01,
        )


@pytest.mark.skipif(
    not has_input_audio_device(),
    reason="No audio device found.",
)
def test_time_expansion_is_saved_in_guano_metadata(
    tmp_path: Path, deployment: data.Deployment, mocker
):
    mocker.patch(
        "acoupi.tasks.recording.get_device_id",
        return_value="1234567890ABCDEF",
    )

    _, samplerate, device_name = get_default_microphone()
    recorder = components.PyAudioRecorder(
        duration=1,
        samplerate=samplerate,
        audio_channels=1,
        device_name=device_name,
        chunksize=8192,
        audio_dir=tmp_path,
        time_expansion=0.1,
    )

    recording = recorder.record(deployment=deployment)

    add_guano_metadata(recording)

    assert recording.path is not None
    g = guano.GuanoFile(str(recording.path))
    assert g["TE"] == str(0.1)
    assert g["Samplerate"] == samplerate
